#encoding: UTF-8 
=begin
/***************************************************************************
 *   ©2018-2019 Michael Uplawski <michael.uplawski@uplawski.eu>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
=end

require_relative 'argparser'
require_relative 'table_template'
require_relative 'styles_template'
require_relative 'logging'
require_relative 'user_input'
require_relative 'file_checking'
require 'mimemagic'
require 'date'
require 'tmpdir'
require 'fileutils'

# The main application class.
# All the work is done upon initialization; if responsibilities were
# better devided, there would also be more OO.
# E.g.: TODO: Separate XML-munging from file-system operations.
# Or : TODO: Separate template manipulation from writing output.

class Generator
  # prepare a class-level logger
  self.extend(Logging)
  @@log = init_logger(STDOUT)
  # probably utterly useless, but I have this around for so long, now. 
  include File_Checking

  # What is whitespace in the XML-code of the final document
  WS = /\n\s*</

  DOC = "document.xml"

  def initialize(args)
    @log = @@log
    if args.length < 1
      @log.error("No Arguments. Execute with -h to see an overview.\n\tAborting! Bye.")
      exit false
    end
    @options = ArgParser.parse(args)
    @log.debug('options: ' << @options.inspect)

    # Does it
    generate
  end

  private

  # convenience function  
  def highlight_style(ele, rows)
    if(@options.highlight && @options.date && ( ele.wday == 6 || ele.wday == 0)) 
      # add a paragraph-style
      rows << '<w:rPr><w:b/><w:bCs/></w:rPr>'
    end
  end

  # convenience function
  def week_style(ele, rows)
    if(@options.weeks && @options.date && ele.wday ==  1)
      # add a cell-style
      rows << '<w:tcPr><w:tcBorders>'
      rows << '<w:top w:val="single" w:sz="16" w:space="0" w:color="000000" tmln="40, 20, 20, 0,0"/>'
      rows << '</w:tcBorders></w:tcPr>'
    end
  end

  # Convenience function.
  # Remove stale files and folders from the 
  # word-folder, then the word-folder itself.
  def cleanup
    if Dir.exist?(@word)
      begin
        Dir.each_child(@word) do |ele| 
          fl = File.join(@word,ele)
          if File.file?(fl)
            File.delete(fl) 
          elsif File.directory?(fl)
            Dir.each_child(fl) do |e| 
              fle = File.join(fl, e) 
              @log.debug('removing file ' << fle)
              File.delete(fle) if File.exist?(fle) 
            end
            Dir.rmdir(File.join(@word,ele))
          end
        end
        Dir.rmdir(@word)
      rescue SystemCallError => er
        msg = "\nCannot remove the directory " << @word << " or one of the files therein: " << er.message
        msg << "\nIf you know that the directory is no longer needed, remove it manually, \nthen retry." 
        msg << "\n\tAborting. Bye."
        @log.error(msg)
        exit false
      end
    end
  end

  # Convenience function
  def lookup_templates 
    xml_type = "application/xml"
    if(@options.templates)
      tdir = @options.templates
      templates = Array.new 
      Dir.each_child(tdir) do |fn|
        fmagic = MimeMagic.by_path(fn)
        @log.debug('testing file ' << fn )
        @log.debug('magic is ' << (fmagic ? fmagic.to_s : 'N I L') )
        if(fmagic && fmagic.to_s == xml_type)
          templates << File.join(tdir, fn)
        else
          @log.info('Not a template: ' << fn << (fmagic ? ' (' << fmagic.comment << ')' : '' ) << ' (Ignoring).' ) 
        end
      end
      if templates.length < 2
        @log.error( sprintf "%s does not contain enough template-files:", @options.templates)
        @log.error( sprintf "\t%s", Dir.children(@options.templates).join(', ') )
        @log.error("\tName a folder with at least a document.xml and a corresponding style.xml")
        @log.error("\tor do not name a template folder at all to use the defaults.")
        @log.error("\t ---- Aborting, bye. ----")
        exit false
      end
      rels = File.join(tdir, '_rels')
      if(Dir.exist?(rels) )
        # option 1: return Array + String.
        # Includes headers, footers and the like.
        return templates, rels
      else
        # option 2: return Array of file-names.
        # Flat document & styles
        return templates
      end
    else
      # option 3: return an Array of 2 xml-documents
      # Default when no user-option is given.
      return $TABLE_TEMPLATE.dup, $STYLES_TEMPLATE.dup
    end
  end

  # convenience function
  def from_templates(id)
    @log.debug('templates is ' << @templates.inspect)
    table = @templates[0].respond_to?(:to_ary) ? @templates[0].detect {|ele| ele.end_with?(id)} : nil
    table ||= File.exist?(@templates[0] ) ? @templates.detect{|ele| ele.end_with?(id)} : nil
    if(table)
      table = File.read(table)
    else
      table = @templates[0] if @templates.length == 2 && @templates[0].respond_to?(:to_str)
    end
    if(!table) 
      @log.error("Cannot identify the template for %s.\n\tAborting. Bye!" %id)
      exit false
    end
    return table
  end

  # derive the mime-type from the file
  def self::file_type(fl)
    fm = MimeMagic.by(File.new(fl))
  end

  # A monster of a function. But I have seen worse, especially
  # in the context of XML-generation. 
  def generate
    sequence = @options.date ? date_sequence : number_sequence 

    @templates = lookup_templates() 
    table = from_templates(DOC )
    rows = ''
    # Create content
    sequence.each do |ele|
      # Do not generate rows, if week-ends must be excluded
      if(@options.weekend || @options.date && ele.wday != 6 && ele.wday != 0)
        rows << '<w:tr><w:tc>'
        # delimit weeks?
        week_style(ele, rows)
        rows << '<w:p><w:r>'
        # highlight week-ends?
        highlight_style(ele, rows)
        rows << '<w:t>'
        rows << format(ele) << '</w:t></w:r></w:p></w:tc>'
        # empty cells
        (@options.cols - 1).times do 
          rows << '<w:tc>'
          week_style(ele, rows)
          rows << '</w:tc>' 
        end
        rows << '</w:tr>'
      end
    end
    @log.debug('rows is ' << rows)
    # TODO: Whatever the future of this thing, remember to
    # set the log-level to info.
    table.sub!('<w:tr></w:tr>', rows.to_s)
    if !@options.file
      # No file wanted: show the result and exit
      @log.info(DOC.dup << ":\n" << table.inspect)
      exit true
    elsif File.exist?(@options.file)
      @log.warn(@options.file.dup << " exists !\n\tShould the file be replaced (y/N)?")
      if wait_for_user.chr.downcase != 'y'
        @log.info('Okay, doing nothing. Bye.')
        exit true
      end
    end
    # The remainder serves to write or copy the xml-code and to create an
    # archive of the given output-filename.

    # Locate the future word folder in the system-wide temp-folder.
    # Remove the remainder of previous attempts and recreate a new and clean
    # word-folder.
    @word = File.join(Dir::tmpdir, 'word')
    cleanup 
    Dir.mkdir(@word) 
    # Locate and write the new files.
    File.write(File.join(@word, DOC), table.gsub(WS,'<'))
    tpl_ary = @templates[0].respond_to?(:to_str) ? @templates : @templates[0]
    tpl_ary.each do |tpl|
      if(File.file?(tpl) && !tpl.end_with?(DOC))
        # Upon copying the xml-content, remove unneeded white-space.
        # ATTN! The difference in size can be considerable between final 
        # TMDX-files.
        File.write(File.join(@word, File.basename(tpl)), File.read(tpl).gsub(WS,'<'))
      end
    end
    # If need be, also create the _rels subfolder and its content.
    rels_tpl = @templates.detect{|ele| ele.respond_to?(:to_str) && File.basename(ele) == '_rels' }
    if(rels_tpl) 
      rels = File.join(@word, '_rels')
      Dir.mkdir(rels )
      # Note that this will copy not more than 1 file, most of the time.
      Dir.each_child(rels_tpl) {|fl| File.write(File.join(rels, fl), File.read(File.join(rels_tpl, fl))) }
    end
    Dir::chdir(Dir::tmpdir)
    File.delete(@options.file) if File.exist?(@options.file)
    # TODO: Use the Escape gem.
    # This compresses the word-folder and stores the result under the name
    # given on the command-line.
    #
    # I hope that it does not do other things, as well. 
    msg = 'Output created by zip:'
    @log.info(msg )
    @log.info('▔' * msg.length)
    thr = nil
    begin
      # TODO: Find out, why the program exits after the execution of the
      # following command.
      thr = Thread.new {exec("zip", "-9r", "--verbose", @options.file, File.basename(@word) )}
    rescue Exception => ex
      @log.error("Cannot compress the word-folder: " << ex.message)
      exit false
    end
    thr.join
    # Verify
    if File.exist?(@options.file) 
      @log.info(@options.file << " has been written")
      # Clean away the current word-folder
      cleanup
    else
      @log.error(@options.file << " could NOT be written.")
      @log.error(@word << " is still available for checks.")
    end
  end

  # Formats dates or numbers as imposed by the command-line-options. 
  def format(ele)
    ele.respond_to?(:to_date) ? ele.strftime(@options.format) : sprintf("%" << @options.n_width.to_s << "d", ele)
  end

  # creates a sequence of dates
  def date_sequence
    sd = Date.parse(@options.s_value)
    ed = Date.parse(@options.e_value)
    seq = Array.new
    sd.upto(ed) do |dt| 
      seq << dt
    end
    seq
  end

  # Creates a sequence of numbers.
  # This is not the original purpose of the program
  # and most of the functionality makes sense only
  # with date-sequences.
  # TODO: maybe remove the number-sequence; also from 
  # the options.
  def number_sequence
    seq = Array.new
    @options.s_value.upto(@options.e_value) {|n| seq << n}
    seq
  end
end
# EOF
